/*
 * Big Brother memory monitor external for Windows.
 *
 * Based on bb_memory 2.0.1 from www.deadcat.net, with
 * modifications to use new GlobalMemorystatusEx API
 * and obtain some option-values from the BB registry
 * configuration.
 *
 * bb_memory 2.0.1 is:
 * "Copyright (C) 2000 Bennett Technologies.      All rights reserved"
 * (see the original license in the LICENSE.BEN)
 *
 * Redistribution of this modified version has been agreed with
 * Bennett Samowitch by e-mail, Feb 12 2004.
 *
 * The sourcefile upon which this is based is
 * "$Id: bb_memory.c,v 3.3 2004/09/28 16:42:28 henrik Exp $"
 *
 * Development after the release of bb_memory 2.0.1 is
 * (C) Henrik Strner <henrik-bb@hswn.dk>
 *
 */

static char VERSION[] = "$Id: bb_memory.c,v 3.3 2004/09/28 16:42:28 henrik Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <windows.h>
#include <winsock.h>
#include <conio.h>

/* Version information */

/* Maximum length of a hostname.  Needed for gethostname() */
#define MAXHOSTNAMELEN	64

/*
 * This will become a column on the DISPLAY.  It should
 * be as short as possible to save space...
 * Note: You can also create a help file for this test
 * which should be put in www/help/$LABEL_MEM.html.  It will
 * be linked into the display automatically.
 */
#define LABEL_MEM	"memory"
#define LABEL_NET	"netstat"

/*
 * Set some default values.  We will go "yellow" at WARN
 * and "red" at PANIC.
 */
#define DEF_PHYS_WARN	100		/* physical memory */
#define DEF_PHYS_PANIC	101
#define DEF_VIRT_WARN	90		/* virtual memory */
#define DEF_VIRT_PANIC	99
#define DEF_PAGE_WARN	80		/* page-file (swap) */
#define DEF_PAGE_PANIC	90

char *BB_COLORS[] = { "clear", "green", "yellow", "red", "purple", "blue" };

#define BB_CLEAR	0
#define BB_GREEN	1
#define BB_YELLOW	2
#define BB_RED		3
#define BB_PURPLE	4
#define BB_BLUE		5

struct level {
    unsigned long value;
    unsigned long total;
    unsigned long used;
    unsigned int  warn;
    unsigned int  panic;
    unsigned int  color;
};

char *pname;
char *path = NULL;
int debug = 0;
int donetstat = 1;

/* For dynamic linking of GlobalMemoryStatus* functions */
HINSTANCE hinstLib;
typedef BOOL(*MYPROCEX) (LPMEMORYSTATUSEX);
typedef BOOL(*MYPROC) (LPMEMORYSTATUS);
MYPROCEX  procMemStatusEx;
MYPROC    procMemStatus;


void usage(void);
void split(char*, struct level*);
void setcolor(struct level*);
void swpchr(char*, char, char);

int main(int argc, char **argv)
{
    MEMORYSTATUSEX 	Memory;
    WSADATA 	 	wsaData;
    struct level 	phys;
    struct level 	virt;
    struct level 	page;
    struct hostent 	*he;
    char                szLoglocation[2048];
    char                *filename;
    char                *tmpfilename;
    char         	host[MAXHOSTNAMELEN + 1];
    int 		pagecolor;
    char		*status;
    int          	err;
    int			usefqdn = 0;
    time_t		t;
    char		*date;
    FILE		*fp;
    int                 haveData = 0;

    /* Start WINSOCK.DLL */
    err = WSAStartup( MAKEWORD( 1, 1 ), &wsaData );
    if ( err != 0 ) {
	/* Tell the user that we couldn't find a usable WinSock DLL. */
    	printf ("WSAStartup() error\n");
	exit ( -1 );
    }

    /*
     * Get/Set the name of this application.  Windows
     * likes to prepend the entire path, so we must
     * remove it.
     */
    pname = argv[0];
    if (strrchr(argv[0], '\\') != NULL) {
       pname = (char*)(strrchr(argv[0], '\\')) + 1;
    }

    /* Initialize some default values */
    phys.warn  = DEF_PHYS_WARN;  	/* physical memory */
    phys.panic = DEF_PHYS_PANIC;
    virt.warn  = DEF_VIRT_WARN;		/* virtual memory */
    virt.panic = DEF_VIRT_PANIC;
    page.warn  = DEF_PAGE_WARN;		/* page-file (swap) */
    page.panic = DEF_PAGE_PANIC;

    /*
     * Setup the default path for the logfile. This is taken
     * from the BBNT client "Saved Logs" registry setting.
     */
    {
       HKEY hKey;
       DWORD dwBuflen;
       LONG lRet;

       lRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE, 
                "SOFTWARE\\Quest Software\\BigBrother\\bbnt\\ExternalPath",
                0, KEY_QUERY_VALUE, &hKey);

       if (lRet != ERROR_SUCCESS) {
          /* Try the old location */
          lRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE, 
                "SOFTWARE\\BigBrother\\bbnt\\ExternalPath",
                0, KEY_QUERY_VALUE, &hKey);
       }

       if (lRet == ERROR_SUCCESS) {
          dwBuflen = sizeof(szLoglocation);
          lRet = RegQueryValueEx(hKey, NULL, NULL, NULL, (LPBYTE)szLoglocation, &dwBuflen);
       }

       if (lRet == ERROR_SUCCESS) {
          path = szLoglocation;
       }

       RegCloseKey(hKey);
    }

    while (*++argv != NULL && **argv == '-') {
    	switch (argv[0][1]) {
       	    case 'F':	/* use FQDN */
            	usefqdn = 1;
                break;
            case 'P':	/* Set the physical memory ranges */
            	if (*++argv == NULL)
                	usage();
                split(*argv, &phys);
                break;
            case 'V':	/* Set the virtual memory ranges */
            	if (*++argv == NULL)
                	usage();
                split(*argv, &virt);
                break;
            case 'S':	/* Set the page-file (swap) ranges */
            	if (*++argv == NULL)
                	usage();
                split(*argv, &page);
                break;
            case 'p':	/* Set the path to place the output */
            	if (*++argv == NULL) path = NULL;
                else path = *argv;
                break;
            case 'h':	/* Help me, I'm on the loose :) */
            	usage();
                exit (1);	/* should never get here */
            	break;
	    case 'd':
		debug = 1;
		break;
            default:
            	(void)fprintf (stderr, "unknown option: %s\n", *argv);
                usage();
                exit (1);	/* should never get here */
                break;
 	}
    }

    if (debug && path) printf("Logfile path found: %s\n", path);

    /*
     * Find a GlobalMemoryStatus function that works. We
     * prefer GlobalMemoryStatusEx, but this is available 
     * only on Windows 2000/XP.
     * So load the kernel32 DLL and see what function we have.
     */
    hinstLib = LoadLibrary("kernel32.dll");

    if (hinstLib != NULL) {
       procMemStatusEx = (MYPROCEX)GetProcAddress(hinstLib, "GlobalMemoryStatusEx");
       if (procMemStatusEx != NULL) {
	  if (debug) printf("Using new GlobalMemoryStatusEx API\n");
          Memory.dwLength = sizeof(Memory);
          (procMemStatusEx)(&Memory);
          haveData = 1;
       }
    }

    if (!haveData) {
       if (hinstLib != NULL) {
          procMemStatus = (MYPROC)GetProcAddress(hinstLib, "GlobalMemoryStatus");
          if (procMemStatus != NULL) {
             MEMORYSTATUS M1;
	     if (debug) printf("Using old GlobalMemoryStatus API\n");
             M1.dwLength = sizeof(M1);
             (procMemStatus)(&M1);
             Memory.dwLength          = M1.dwLength;
             Memory.dwMemoryLoad      = 0;
             Memory.ullTotalPhys      = M1.dwTotalPhys;
             Memory.ullAvailPhys      = M1.dwAvailPhys;
             Memory.ullTotalPageFile  = M1.dwTotalPageFile;
             Memory.ullAvailPageFile  = M1.dwAvailPageFile;
             Memory.ullTotalVirtual   = M1.dwTotalVirtual;
             Memory.ullAvailVirtual   = M1.dwAvailVirtual;
             haveData = 2;
          }
       }
    }

    if (hinstLib) FreeLibrary(hinstLib);

    if (!haveData) {
       printf("FATAL: Cannot find a GlobalMemoryStatus function\n");
       return 1;
    }

    if (debug) {
       printf("TotalPhys     = %llu\n", Memory.ullTotalPhys);
       printf("AvailPhys     = %llu\n", Memory.ullAvailPhys);
       printf("TotalPageFile = %llu\n", Memory.ullTotalPageFile);
       printf("AvailPageFile = %llu\n", Memory.ullAvailPageFile);
       printf("TotalVirtual  = %llu\n", Memory.ullTotalVirtual);
       printf("AvailVirtual  = %llu\n", Memory.ullAvailVirtual);
       exit(0);
    }

    /* Convert to MB to avoid overflowing later calculations */
    Memory.ullTotalPhys     /= (1024*1024);
    Memory.ullAvailPhys     /= (1024*1024);
    Memory.ullTotalPageFile /= (1024*1024);
    Memory.ullAvailPageFile /= (1024*1024);
    Memory.ullTotalVirtual  /= (1024*1024);
    Memory.ullAvailVirtual  /= (1024*1024);

    /* Calculate the usage percentage */
    phys.total = Memory.ullTotalPhys;
    phys.used  = (Memory.ullTotalPhys - Memory.ullAvailPhys);
    phys.value = (phys.used * 100) / Memory.ullTotalPhys;

    virt.total = Memory.ullTotalVirtual;
    virt.used  = (Memory.ullTotalVirtual - Memory.ullAvailVirtual);
    virt.value = (virt.used * 100) / Memory.ullTotalVirtual;

    page.total = Memory.ullTotalPageFile;
    page.used  = (Memory.ullTotalPageFile - Memory.ullAvailPageFile);
    page.value = (page.used * 100) / Memory.ullTotalPageFile;

    /* Set the individual line colors */
    setcolor(&phys);
    setcolor(&virt);
    setcolor(&page);

    pagecolor = BB_GREEN;
    status    = "Memory OK";

    if (phys.color == BB_RED) {
    	pagecolor = BB_RED;
    	status    = "Memory **very** low";
    } else if (phys.color == BB_YELLOW) {
    	if (pagecolor != BB_RED) {
    		pagecolor = BB_YELLOW;
        	status    = "Memory low";
        }
    }
    if (virt.color == BB_RED) {
    	pagecolor = BB_RED;
    	status    = "Memory **very** low";
    } else if (virt.color == BB_YELLOW) {
    	if (pagecolor != BB_RED) {
    		pagecolor = BB_YELLOW;
        	status    = "Memory low";
        }
    }
    if (page.color == BB_RED) {
    	pagecolor = BB_RED;
    	status    = "Memory **very** low";
    } else if (page.color == BB_YELLOW) {
    	if (pagecolor != BB_RED) {
    		pagecolor = BB_YELLOW;
        	status    = "Memory low";
        }
    }

    /* Get this machine's name.  Convert to FQDN if requested. */
    gethostname((char*)&host, sizeof(host));
    if (usefqdn) {
    	if ((he = gethostbyname(host)) != NULL) {
        	strncpy(host, he->h_name, sizeof(host));
        }
    }

    /*
     * Get the current time.  Since we are using ctime(), we
     * must lose the '\n' from the returned string as well.
     */
    time (&t);
    date = ctime(&t);
    *((char*)(strrchr(date, '\n'))) = '\0';

    /* Display the output */
    fp = stdout;
    if (!debug && (path != NULL)) {
    	filename = (char*)malloc(strlen(path) + strlen(LABEL_MEM) + 5);
    	tmpfilename = (char*)malloc(strlen(path) + strlen(LABEL_MEM) + 9);

    	sprintf (filename, "%s\\%s", path, LABEL_MEM);
    	sprintf (tmpfilename, "%s\\%s.tmp", path, LABEL_MEM);
    	fp = fopen(tmpfilename, "w");
    }

    fprintf (fp, "%s %s [%s]\n", BB_COLORS[pagecolor], date, host);
    swpchr((char*)&host, '.', ',');
    fprintf (fp, "status %s.%s %s %s - %s\n", host, LABEL_MEM,
    	BB_COLORS[pagecolor],  date, status);
    fprintf (fp, "    Memory  %6s  %6s  %s\n", "Used", "Total", "Pctg");
    fprintf (fp, "&%s Physical:  %6luM %6luM  %3lu%%\n", BB_COLORS[phys.color],
    	phys.used, phys.total, phys.value);
    fprintf (fp, "&%s Virtual:   %6luM %6luM  %3lu%%\n", BB_COLORS[virt.color],
    	virt.used, virt.total, virt.value);
    fprintf (fp, "&%s Page:      %6luM %6luM  %3lu%%\n", BB_COLORS[page.color],
    	page.used, page.total, page.value);
    if (fp == stdout) fprintf (fp, "\n<FONT SIZE=\"-1\">%s</FONT>\n", VERSION);

    if (path != NULL) {
       fclose(fp);
       rename(tmpfilename, filename);
    }

    if (donetstat) {
        /* Do the netstat stuff */
	FILE *nsfp;
	int nsreturncode;

        nsreturncode = system("netstat -s >c:\\netstat.tmp");
	nsfp = fopen("c:\\netstat.tmp", "r");
	if (nsfp && (nsreturncode == 0)) {
	    char buf[512];
	    char *prefix = "";

	    fp = stdout;
	    if (!debug && (path != NULL)) {
	    	filename = (char*)malloc(strlen(path) + strlen(LABEL_NET) + 5);
	    	tmpfilename = (char*)malloc(strlen(path) + strlen(LABEL_NET) + 9);

	    	sprintf (filename, "%s\\%s", path, LABEL_NET);
	    	sprintf (tmpfilename, "%s\\%s.tmp", path, LABEL_NET);
	    	fp = fopen(tmpfilename, "w");
	    }

	    fprintf (fp, "%s %s [%s]\n", BB_COLORS[BB_GREEN], date, host);
	    swpchr((char*)&host, '.', ',');
	    fprintf (fp, "win32\n");

	    while (fgets(buf, sizeof(buf), nsfp)) {
	            if (strncmp(buf, "IP Statistics", 13) == 0) {
			prefix = "ip";
			fprintf(fp, "\n");
		    }
		    else if (strncmp(buf, "ICMP Statistics", 15) == 0) {
			prefix = "icmp";
			fprintf(fp, "\n");
		    }
		    else if (strncmp(buf, "TCP Statistics", 14) == 0) {
			prefix = "tcp";
			fprintf(fp, "\n");
		    }
		    else if (strncmp(buf, "UDP Statistics", 14) == 0) {
			prefix = "udp";
			fprintf(fp, "\n");
		    }
		    else if (strchr(buf, '=')) {
			char *p;
			p = strtok(buf, " \t");
			fprintf(fp, "%s", prefix);
			while (p) {
				fprintf(fp, "%s", p);
				p = strtok(NULL, " \t");
			}
		    }
	    }
	    fclose(nsfp);
	}
	unlink("c:\\netstat.tmp");
    }

    if (path != NULL) {
       fclose(fp);
       rename(tmpfilename, filename);
    }

    return 0;
}

/*
 * swap all occurances of a character with another.
 */
void swpchr(char *original, char bad, char good)
{
        int     i;

        i = 0;
        while (original[i]) {
                if (original[i] == bad)
                        original[i] = good;
                i++;
        }
}

void setcolor(struct level *level)
{
    level->color = BB_GREEN;

    if (level->value >= level->warn)
    	level->color = BB_YELLOW;

    if (level->value >= level->panic)
    	level->color = BB_RED;
}

void split(char *arg, struct level *level)
{
    char *value;

    if ((value = strtok (arg," ")) == NULL)
    	usage();
    level->warn  = atoi(value);

    if ((value = strtok (NULL," ")) == NULL)
    	usage();
    level->panic  = atoi(value);
}

void usage()
{
	(void)printf ("usage: %s [ OPTIONS ]\n", pname);
    (void)printf ("where:\n");
    (void)printf ("   -h             \tPrint this message.\n");
    (void)printf ("   -p <PATH>      \tSet the PATH to the output.\n");
    (void)printf ("                  \t(creates the file: <PATH>\\%s)\n", LABEL_MEM);
    (void)printf ("                  \tDefault: %s\n", (path ? path : "NONE"));
    (void)printf ("                  \tIf empty: Show on stdout\n");
    (void)printf ("   -P <WARN PANIC>\tSet WARN and PANIC levels for physical memory (%d:%d)\n", 
                  DEF_PHYS_WARN, DEF_PHYS_PANIC);
    (void)printf ("   -S <WARN PANIC>\tSet WARN and PANIC levels for paging (swap) memory (%d:%d)\n",
                  DEF_PAGE_WARN, DEF_PAGE_PANIC);
    (void)printf ("   -V <WARN PANIC>\tSet WARN and PANIC levels for virtual memory (%d:%d)\n",
                  DEF_VIRT_WARN, DEF_VIRT_PANIC);
    (void)printf ("\n");
    (void)printf ("\nNote: All arguments must be provided.\n");
    exit (1);
}

